@@ -19,10 +19,6 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#include <linux/config.h>
-#if !CONFIG_PROFILING
-# error Sysprof needs a kernel with profiling support compiled in.
-#endif
 #ifdef CONFIG_SMP
 # define __SMP__
 #endif
@@ -43,7 +39,16 @@
 #include "../config.h"
 
 #include <linux/version.h>
-#if KERNEL_VERSION(2,6,11) > LINUX_VERSION_CODE
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
+#include <linux/config.h>
+#endif
+
+#if !CONFIG_PROFILING
+# error Sysprof needs a kernel with profiling support compiled in.
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
 # error Sysprof needs a Linux 2.6.11 kernel or later
 #endif
 #include <linux/kallsyms.h>
@@ -111,24 +116,31 @@ read_frame (void *frame_pointer, StackFrame *frame)
 	return 0;
 }
 
+DEFINE_PER_CPU(int, n_samples);
+
 static int
 timer_notify (struct pt_regs *regs)
 {
-	static int n_samples;
 	SysprofStackTrace *trace = head;
 	int i;
 	int is_user;
+	static atomic_t in_timer_notify = ATOMIC_INIT(1);
 
-	if ((n_samples++ % INTERVAL) != 0)
+	if ((++get_cpu_var(n_samples) % INTERVAL) != 0)
 		return 0;
 
+	/* 0: locked, 1: unlocked */
+	
+	if (!atomic_dec_and_test(&in_timer_notify))
+		goto out;
+	
 	is_user = user_mode(regs);
 
 	if (!current || current->pid == 0)
-		return 0;
+		goto out;
 	
 	if (is_user && current->state != TASK_RUNNING)
-		return 0;
+		goto out;
 
 	if (!is_user)
 	{
@@ -176,7 +188,9 @@ timer_notify (struct pt_regs *regs)
 		head = &stack_traces[0];
 	
 	wake_up (&wait_for_trace);
-	
+
+out:
+	atomic_inc(&in_timer_notify);
 	return 0;
 }
 
@@ -192,6 +206,8 @@ procfile_read(char *buffer,
 		return -EWOULDBLOCK;
 	
 	*buffer_location = (char *)tail;
+
+	BUG_ON(tail->pid == 0);
 	
 	if (tail++ == &stack_traces[N_TRACES - 1])
 		tail = &stack_traces[0];
